package eu.mhsl.minenet.minigames.instance.game.stateless.types.towerdefense;

import eu.mhsl.minenet.minigames.instance.Dimension;
import eu.mhsl.minenet.minigames.instance.game.stateless.types.towerdefense.generator.MazeGenerator;
import eu.mhsl.minenet.minigames.instance.game.stateless.types.towerdefense.towers.Tower;
import eu.mhsl.minenet.minigames.util.BatchUtil;
import eu.mhsl.minenet.minigames.util.CommonEventHandles;
import net.minestom.server.MinecraftServer;
import net.minestom.server.collision.Aerodynamics;
import net.minestom.server.coordinate.Point;
import net.minestom.server.coordinate.Pos;
import net.minestom.server.coordinate.Vec;
import net.minestom.server.entity.*;
import net.minestom.server.entity.attribute.Attribute;
import net.minestom.server.entity.damage.DamageType;
import net.minestom.server.event.entity.EntityDeathEvent;
import net.minestom.server.event.entity.projectile.ProjectileCollideWithBlockEvent;
import net.minestom.server.event.entity.projectile.ProjectileCollideWithEntityEvent;
import net.minestom.server.event.inventory.InventoryPreClickEvent;
import net.minestom.server.event.item.ItemDropEvent;
import net.minestom.server.event.player.*;
import net.minestom.server.instance.InstanceContainer;
import net.minestom.server.instance.block.Block;
import net.minestom.server.item.ItemStack;
import net.minestom.server.item.Material;
import org.jetbrains.annotations.Nullable;

import java.time.Duration;
import java.util.ArrayList;
import java.util.List;
import java.util.UUID;

public class TowerdefenseRoom extends InstanceContainer {
    private final static int reach = 30;
    private long lastPlayerAttack = 0;
    private final Player player;
    private final Towerdefense game;
    private final List<EntityCreature> enemies = new ArrayList<>();
    private final List<Tower> towers = new ArrayList<>();
    private final Entity cursor;
    private int money = 1000;

    public TowerdefenseRoom(Player player, Towerdefense game) {
        super(UUID.randomUUID(), Dimension.OVERWORLD.key);
        MinecraftServer.getInstanceManager().registerInstance(this);
        this.player = player;
        this.game = game;
        this.player.setGameMode(GameMode.ADVENTURE);
        this.player.setAllowFlying(true);
        this.addMoney(0);
        this.player.getAttribute(Attribute.BLOCK_INTERACTION_RANGE).setBaseValue(reach);
        this.player.getAttribute(Attribute.ENTITY_INTERACTION_RANGE).setBaseValue(reach);

        this.player.getInventory().addItemStack(ItemStack.of(Material.SPECTRAL_ARROW));
        this.game.getAvailableTowers().keySet().forEach(material -> this.player.getInventory().addItemStack(ItemStack.of(material)));

        this.setGenerator(new MazeGenerator());
        BatchUtil.loadAndApplyBatch(this.game.getMazeBatch(), this, () -> {});

        this.cursor = new Entity(EntityType.ARMOR_STAND);
        this.cursor.setInstance(this);
        this.cursor.setBoundingBox(0,0,0);

        this.eventNode()
            .addListener(EntityDeathEvent.class, event -> {
                if(!(event.getEntity() instanceof EntityCreature enemy)) return;
                this.enemies.remove(enemy);
            })
            .addListener(PlayerTickEvent.class, this::setCursorPosition)
            .addListener(PlayerUseItemEvent.class, event -> this.useItem())
            .addListener(PlayerUseItemOnBlockEvent.class, event -> this.useItem())
            .addListener(PlayerEntityInteractEvent.class, event -> this.useItem())
            .addListener(PlayerHandAnimationEvent.class, event -> this.useItem())
            .addListener(ItemDropEvent.class, CommonEventHandles::cancel)
            .addListener(InventoryPreClickEvent.class, CommonEventHandles::cancel)
            .addListener(PlayerSwapItemEvent.class, CommonEventHandles::cancel);
    }

    private void useItem() {
        if(this.player.getItemInMainHand().material().equals(Material.SPECTRAL_ARROW)) {
            this.playerAttack();
            return;
        }
        this.placeTower();
    }

    private void playerAttack() {
        if(this.lastPlayerAttack > System.currentTimeMillis()-200) return;
        Player p = this.player;
        this.lastPlayerAttack = System.currentTimeMillis();
        EntityProjectile projectile = new EntityProjectile(p, EntityType.SPECTRAL_ARROW);
        projectile.setView(p.getPosition().yaw(), p.getPosition().pitch());
        projectile.setNoGravity(true);
        projectile.setAerodynamics(new Aerodynamics(0, 1, 0));

        Vec projectileVelocity = p.getPosition().direction().normalize().mul(20);
        projectile.setVelocity(projectileVelocity);
        projectile.scheduleRemove(Duration.ofSeconds(5));
        projectile.setInstance(this, p.getPosition().add(0, p.getEyeHeight(), 0));
        projectile.eventNode()
            .addListener(ProjectileCollideWithEntityEvent.class, hitEvent -> {
                if(!(hitEvent.getTarget() instanceof EntityCreature target)) return;
                if(!this.getEnemies().stream().filter(entityCreature -> !entityCreature.isDead()).toList().contains(target)) return;
                target.damage(DamageType.PLAYER_ATTACK, 0.1f);
                if(target.isDead()) this.addMoney((int) target.getAttribute(Attribute.MAX_HEALTH).getBaseValue());
                projectile.remove();
            })
            .addListener(ProjectileCollideWithBlockEvent.class, collisionEvent -> projectile.remove());
    }

    public void addMoney(int amount) {
        this.money += amount;
        this.player.setLevel(this.money);
    }

    private void placeTower() {
        Material usedMaterial = this.player.getItemInMainHand().material();
        if(!this.canPlaceTower(usedMaterial)) return;
        try {
            Tower tower = this.game.getAvailableTowers().entrySet().stream()
                .filter(materialClassEntry -> materialClassEntry.getKey().equals(usedMaterial))
                .findFirst()
                .orElseThrow()
                .getValue()
                .getConstructor()
                .newInstance();
            this.setBlock(this.cursor.getPosition().sub(0, 0.5, 0), Block.BLUE_WOOL);
            tower.setInstance(this, this.cursor.getPosition());
            this.towers.add(tower);
        } catch (Exception ignored) {
        }
    }

    private void setCursorPosition(PlayerTickEvent event) {
        Point newPosition = this.player.getTargetBlockPosition(reach);
        if(newPosition == null) return;
        newPosition = newPosition.add(0.5,1,0.5);
        this.cursor.teleport(new Pos(newPosition));
        this.cursor.setInvisible(!(this.canPlaceTower((Material) null) && this.game.getAvailableTowers().containsKey(this.player.getItemInMainHand().material())));
    }

    private boolean canPlaceTower(@Nullable Material usedMaterial) {
        if(usedMaterial == null) return this.canPlaceTower((Class<Tower>) null);
        return this.canPlaceTower(this.game.getAvailableTowers().get(usedMaterial));
    }

    private boolean canPlaceTower(@Nullable Class<? extends Tower> towerClass) {
        boolean blockAllowed = this.getBlock(this.cursor.getPosition().sub(0, 1, 0)).equals(Block.BLACK_WOOL);
        if(towerClass == null) return blockAllowed;
        boolean enoughMoney = this.money > this.game.getPrices().get(towerClass);
        return blockAllowed && enoughMoney;
    }

    void startWave(List<GroupFactory> enemyGroups) {
        enemyGroups.forEach(groupFactory -> groupFactory.summonGroup(this));
    }

    public void addEnemy(EntityCreature enemy) {
        enemy.setInstance(this, this.game.getMazePath().getFirst());
        this.enemies.add(enemy);
        this.changeEnemyGoal(enemy, 0);
    }

    private void changeEnemyGoal(EntityCreature enemy, int positionIndex) {
        if(positionIndex == this.game.getMazePath().size()-1) {
            this.enemies.remove(enemy);
            enemy.remove();
            this.player.damage(DamageType.PLAYER_ATTACK, enemy.getHealth()/10);
            return;
        }
        enemy.getNavigator().setPathTo(this.game.getMazePath().get(positionIndex+1), 0.6, () -> this.changeEnemyGoal(enemy, positionIndex+1));
    }

    public Towerdefense getGame() {
        return this.game;
    }

    public List<EntityCreature> getEnemies() {
        return this.enemies;
    }
}
